/*
 * adc_controller.c
 * 
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files 
 * (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

#include "i2c.h"
#include "adc_controller.h"
#include "layout.h"
#include "stdio.h"
#include <avr/pgmspace.h>
#define ADC_ADDRESS 0b1001000

#define CONVERSION_REGISTER_ADDRESS 0x0
#define CONFIG_REGISTER_ADDRESS 0x1

#define SINGLE_SHOT (0b1000000100000000)

#define ADC_MUX_V_IN (0b001<<12)
#define ADC_MUX_V_OUT (0b010<<12)
#define ADC_MUX_V_I (0b000<<12)
#define ADC_MUX_BAT (0b110<<12)
#define ADC_MUX_RANGE_V_IN (0b100<<12) 
#define ADC_MUX_RANGE_V_OUT (0b101<<12)
#define ADC_MUX_RANGE_V_GND (0b111<<12)

#define ADC_GAIN_6V (0b000<<9)
#define ADC_GAIN_4V (0b001<<9)
#define ADC_GAIN_2V (0b010<<9)
#define ADC_GAIN_1V (0b011<<9)
#define ADC_GAIN_05V (0b100<<9)
#define ADC_GAIN_025V (0b101<<9)

#define SPS_128 (0b000<<5)
#define SPS_250 (0b001<<5)
#define SPS_490 (0b010<<5)
#define SPS_920 (0b011<<5)
#define SPS_1600 (0b100<<5)
#define SPS_2400 (0b101<<5)
#define SPS_3300 (0b110<<5)

#define COMP_DISABLE (0b11)

#include <util/delay.h>
#include <math.h>
#include <stdlib.h>

int16_t ADC_single_shot(uint16_t channel){
	I2C_write_word(ADC_ADDRESS,CONFIG_REGISTER_ADDRESS,channel|SINGLE_SHOT|SPS_3300|COMP_DISABLE);
	_delay_us(400);
//	while(!(I2C_read_word(ADC_ADDRESS,CONFIG_REGISTER_ADDRESS))>>15) _delay_us(100);
	return I2C_read_word(ADC_ADDRESS,CONVERSION_REGISTER_ADDRESS);
}

uint16_t const channel_mux_code[7] PROGMEM = {ADC_MUX_V_IN,ADC_MUX_V_OUT,ADC_MUX_V_I,ADC_MUX_BAT,ADC_MUX_RANGE_V_IN,ADC_MUX_RANGE_V_OUT,ADC_MUX_RANGE_V_GND};
uint16_t const range_code[6] PROGMEM = {ADC_GAIN_025V,ADC_GAIN_05V,ADC_GAIN_1V,ADC_GAIN_2V,ADC_GAIN_4V,ADC_GAIN_6V};
	
uint8_t attenuation = 0;
#define ATTENUATION_MULTIPLYER 5.7F
uint8_t ranges[7] = {0};
float const range_multiplyers[6] PROGMEM = {0.256,0.512,1.024,2.048,4.096,6.144};

float adc_measure(uint8_t channel, uint8_t num_times){
	int16_t reading;
	if(attenuation){
		ranges[channel] = 5;
		reading = ADC_single_shot(pgm_read_word(&channel_mux_code[channel])|pgm_read_word(&range_code[ranges[channel]]))/16;
	}
	else for(uint8_t n = 0 ; n < 6 ; n++){
			reading = ADC_single_shot(pgm_read_word(&channel_mux_code[channel])|pgm_read_word(&range_code[ranges[channel]]))/16;
			if(abs(reading)>1900&&ranges[channel]!=6)
				ranges[channel]++;
			else if(abs(reading)<900&&ranges[channel]!=0)
				ranges[channel]--;
			else
				break;
	}
	for(uint8_t n = 1; n < num_times;n++)
		reading += ADC_single_shot(pgm_read_word(&channel_mux_code[channel])|pgm_read_word(&range_code[ranges[channel]]))/16;
	
/*	switch(channel){
		case 0:
			printf("0 %x\n",reading);
		break;
		case 1:
			printf("1 %x\n",reading);
		break;
		case 2:
			printf("2 %x\n",reading);

		break;
	}*/
	
	int16_t offset = eeprom_read_word(&(v_calibration[channel][attenuation][ranges[channel]][1]));
	int16_t linear = eeprom_read_word(&(v_calibration[channel][attenuation][ranges[channel]][0]));
	
//	printf_P(PSTR("mdetud attn %u channel %u range %u %f vrtus %u\n"),attenuation,channel,ranges[channel],(double)pgm_read_float(&range_multiplyers[ranges[channel]]),reading/num_times);
	
	float v_measure = pgm_read_float(&range_multiplyers[ranges[channel]]) * ((((float)linear/16384) * (float)reading/num_times/2048) + (float)offset/16384);
	if(attenuation)
		v_measure *= ATTENUATION_MULTIPLYER;
	return v_measure;	
}

#define PATTERN_SIZE 18 

uint8_t interlaced_pattern[PATTERN_SIZE] = {ADC_CHANNEL_V_IN,ADC_CHANNEL_V_OUT,ADC_CHANNEL_V_DIFF,
											ADC_CHANNEL_V_OUT,ADC_CHANNEL_V_DIFF,ADC_CHANNEL_V_IN,
											ADC_CHANNEL_V_DIFF,ADC_CHANNEL_V_IN,ADC_CHANNEL_V_OUT,
											ADC_CHANNEL_V_IN,ADC_CHANNEL_V_DIFF,ADC_CHANNEL_V_OUT,
											ADC_CHANNEL_V_DIFF,ADC_CHANNEL_V_OUT,ADC_CHANNEL_V_IN,
											ADC_CHANNEL_V_OUT,ADC_CHANNEL_V_IN,ADC_CHANNEL_V_DIFF};
uint8_t interlaced_pattern_index = 0; 

//measures v_in, v_out and v_dif num_times switching between after every measurment
void adc_measure_interlaced(float result[3], float adc_step[3], uint8_t num_times){
	result[0] = result[1] =	result[2] = 0;
	
	for(uint8_t n = 0;n<num_times*3;n++){
		uint8_t ch = interlaced_pattern[interlaced_pattern_index];
		interlaced_pattern_index = (interlaced_pattern_index + 1) % PATTERN_SIZE;
		result[ch] += adc_measure(ch, 1);

	}
	
	adc_step[0] = pgm_read_float(&range_multiplyers[ranges[ADC_CHANNEL_V_IN]]);
	adc_step[1] = pgm_read_float(&range_multiplyers[ranges[ADC_CHANNEL_V_OUT]]);
	adc_step[2] = pgm_read_float(&range_multiplyers[ranges[ADC_CHANNEL_V_DIFF]]);

	for(uint8_t n = 0;n<3;n++){
		result[n]/=num_times;
		adc_step[n]/=2048;
		if(attenuation)
			adc_step[n] *= ATTENUATION_MULTIPLYER;
	}
	
	if(fabs(result[0])>2||fabs(result[1])>2||fabs(result[2])>2){
		attenuation = 1;
		ATTN_SW(1);
	}
	else if(fabs(result[0])<1.5&&fabs(result[1])<1.5&&fabs(result[2])<1.5){
		attenuation = 0;
		ATTN_SW(0);
	}
}

/*
//measures v_in, v_out and v_dif num_times one at a time
void adc_measure_block(float result[3], float adc_step[3], uint8_t num_times){
	result[0] = 0;
	result[1] = 0;
	result[2] = 0;
	
	for(uint8_t n = 0;n<num_times;n++)
		result[0] += adc_measure(ADC_CHANNEL_V_IN, 1);
	for(uint8_t n = 0;n<num_times;n++)
		result[1] += adc_measure(ADC_CHANNEL_V_OUT,1);
	for(uint8_t n = 0;n<num_times;n++)
		result[2] += adc_measure(ADC_CHANNEL_V_DIFF,1);
	
	
	adc_step[0] = pgm_read_float(&range_multiplyers[ranges[ADC_CHANNEL_V_IN]]);
	adc_step[1] = pgm_read_float(&range_multiplyers[ranges[ADC_CHANNEL_V_OUT]]);
	adc_step[2] = pgm_read_float(&range_multiplyers[ranges[ADC_CHANNEL_V_DIFF]]);
	for(uint8_t n = 0;n<3;n++){
		result[n]/=num_times;
		adc_step[n]/=2048;
		if(attenuation)
			adc_step[n] *= ATTENUATION_MULTIPLYER;
	}
}
*/
/*
//check for attenuation needs and attenuate
void attenuate(){
	float v_in = fabs(adc_measure(ADC_CHANNEL_V_IN_RANGE,1));
	float v_out = fabs(adc_measure(ADC_CHANNEL_V_OUT_RANGE,1));
	float v_gnd = fabs(adc_measure(ADC_CHANNEL_V_GND_RANGE,1));
	if(v_in>1.4 || v_out>1.4 || v_gnd>1.4){
		attenuation = 1;
		ATTN_SW(1);
	}		
	else if(v_in<1.1 && v_out<1.1 && v_gnd<1.1){
		attenuation = 0;
		ATTN_SW(0);
	}		
}
*/
