/*
 * filter.c
 * 
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files 
 * (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */
 
#include <stdio.h>
#include <math.h>
#include <avr/pgmspace.h>

#include "filter.h"
#include "layout.h"
#include "adc_controller.h"

#define FILTER_MAX_HISTORY 20
#define FILTER_MIN_AVERAGE 5
#define NUM_SAMPLES_PER_MEASUREMENT 18

float history[2][FILTER_MAX_HISTORY]; //[v_in/v_out/v_diff]
uint8_t history_index;

struct filtered_data latest_data;

//initializes filter when starting measurements
void filter_take_first_data(void){
	attenuate();
	float result[3];
	float adc_step[3];
	adc_measure_interlaced(result,adc_step, NUM_SAMPLES_PER_MEASUREMENT);
	for(uint8_t i = 0; i < FILTER_MAX_HISTORY; i++){
		history[0][i] = result[0]; 
		history[1][i] = result[1];
		history[2][i] = result[2];
	}
	history_index = 0;
	float resistor = eeprom_read_float(&i_resistor_calibration);
	latest_data.current = result[2]/resistor;
	latest_data.current_der = 0;
	latest_data.voltage = (result[1] + result[0] - result[2])/2;
	latest_data.voltage_der = 0;
	latest_data.time = time_now();
}

//updates filter with new data from adc
void filter_take_new_data(void){
	//TODO attenuate and optimize
	
	//attenuate();
	float adc_step[3];
	float result[3];
	adc_measure_interlaced(result,adc_step, NUM_SAMPLES_PER_MEASUREMENT);
	
	uint32_t time_buf = latest_data.time;
	latest_data.time = time_now();
	
	history_index = (history_index + 1) % FILTER_MAX_HISTORY;

	history[0][history_index] = (result[0] + result[1] - result[2])/2; 
	history[1][history_index] = result[2];
	
	float resistor = eeprom_read_float(&i_resistor_calibration);

	for(uint8_t i = 0; i < 2; i++){
		float xm = 0;
        float ym = 0;
		float yym = 0;
		float xym = 0;
		float xxm = 0;
		
		float step = adc_step[i+1]*adc_step[i+1] * 0.3;
		adc_step[i+1] = NAN;
		float min_e = step;
		
		for(uint8_t j = 0; j < FILTER_MAX_HISTORY; j++){
			uint8_t hist_i = (FILTER_MAX_HISTORY + history_index - j)%FILTER_MAX_HISTORY;
			xm += j;
			xxm += ((float)j)*j;
			float dy = history[i][hist_i] - history[i][history_index];
			ym += dy;
			yym += dy*dy;
			xym += j*dy;
			
			if(j >= FILTER_MIN_AVERAGE){
				float b = (xym*(j+1) - xm*ym)/(xxm*(j+1) - xm*xm);
				float a = (ym - b*xm)/(j+1);
				float e = (yym + a*a*(j+1) + b*xxm*b - 2*a*ym - 2*b*xym + 2*a*b*xm)/j  + step*(FILTER_MAX_HISTORY - j)/(10.F*FILTER_MAX_HISTORY) ;
				
				//if(i == 0)
				//	printf("e:%e s:%e\n",e*1000,step*1000);

				if(e > step)
					break;
				if(e < min_e){
					min_e = e;
					result[i+1] = a + history[i][history_index];
					adc_step[i+1] = b;
//					printf("%u a:%f\n",i,a);
				}
			}
		}
//		printf("mine %g %g\n",min_e,step);
	}
	

	float dt = (latest_data.time - (float)time_buf)/SECOND_TO_TIME;
//	printf("dt %f\n",dt);
	latest_data.voltage = result[1];
	latest_data.current = result[2]/resistor;
		
	latest_data.voltage_der = -adc_step[1]/(0.16 * 6/5);
	latest_data.current_der = -adc_step[2]/(resistor*(0.16 * 6/5));
}
