/*
 * i2c.c
 * Based on USB AVR Si570 firmware
 * This software is freely available for non-commercial 
 * use - i.e. for research and experimentation only!
 * 
 * Authors: OBJECTIVE DEVELOPMENT Software GmbH
 *          Christian Starkjohann, F.W. Krom, PE0FKO
 *          Tom Baier DG8SAQ and Hedgehog LLC
 * 
 */

#include <avr/io.h>
#include <util/delay.h>
#include "i2c.h"
#include "layout.h"

#define	I2C_WRITE_BIT		0
#define	I2C_READ_BIT		1

#define I2C_SDA_LO			I2C_DDR |= _BV(I2C_SDA)
#define I2C_SDA_HI			I2C_DDR &= ~_BV(I2C_SDA)
#define I2C_SCL_LO			I2C_DDR |= _BV(I2C_SCL)
#define I2C_SCL_HI			I2C_DDR &= ~_BV(I2C_SCL)

void	I2CSendStart(void);
void	I2CSendStop(void);
void	I2CSendByte(uint8_t b);
void 	I2CSend0(void);
void 	I2CSend1(void);
uint8_t	I2CReceiveByte(void);
void	I2CStretch(void);



uint8_t	I2CError;

void I2CStretch(void)							// Wait until clock hi
{										// Terminate the loop after max 5/bitrate
	uint16_t i = 500; 
	do {
		I2C_DELAY();						// Delay some time
		if (i-- == 0)
		{
			I2CError = 1;			// Error timeout
			break;
		}
	}
	while(!(I2C_PIN & _BV(I2C_SCL)));			// Clock line still low
}

/*
 * Generates a start condition on the bus.
 *
 *	SDA: ..\____..
 *	       __
 *	SCL: ..  \__.. 
 */
void I2CSendStart(void)
{
	I2CError = 0;					// reset error flag
	I2C_SCL_HI;		I2C_DELAY();
	I2C_SDA_LO;  	I2C_DELAY(); 		// Start SDA to low
	I2C_SCL_LO;  	I2C_DELAY();			// and the clock low
}

/*
 * Generates a stop condition on the bus.
 *	           __
 *	SDA: ..___/  ..
 *	        _____
 *	SCL: ../     .. 
 */
void I2CSendStop(void)
{
	I2C_SDA_LO;		I2C_DELAY();
	I2C_SCL_HI;		I2C_DELAY();
	I2C_SDA_HI;		I2C_DELAY();
}

void I2CSend0(void)
{
	I2C_SDA_LO;		I2C_DELAY();		// Data low = 0
	I2C_SCL_HI;		I2CStretch();
	I2C_SCL_LO;		I2C_DELAY();
}

void I2CSend1(void)
{
	I2C_SDA_HI;		I2C_DELAY();		// Data high = 1
	I2C_SCL_HI;		I2CStretch();
	I2C_SCL_LO;		I2C_DELAY();
}

static uint8_t I2CGetBit(void)
{
	uint8_t b;
	I2C_SDA_HI;		I2C_DELAY();		// Data high = input (opencollector)
	I2C_SCL_HI;		I2CStretch();		// SDA Hi Z and wait
	b = (I2C_PIN & _BV(I2C_SDA));				// get bit
	I2C_SCL_LO;		I2C_DELAY();							// clock low
	return b;
}

void I2CSendByte(uint8_t b)
{
	uint8_t i,p;
	p = 0x80;
    for (i=0; i<8; i++)
	{
		if ((p & b) == 0) I2CSend0(); else I2CSend1();
    	p = p >> 1;
	};
    I2CError |= I2CGetBit();	 		//Acknowledge
  	return; 
}

uint8_t I2CReceiveByte(void)
{
	uint8_t i;
	uint8_t b = 0;
    for (i=0; i<8; i++)
	{
		b = b << 1;
		if (I2CGetBit()) b |= 1;
  	};
  	return b;
}

uint8_t	 I2C_read_byte(uint8_t slave, uint8_t register_address){
	I2CSendStart();
	I2CSendByte((slave<<1)|I2C_WRITE_BIT);
	
	if(I2CError){
		I2CSendStop();
		return 0;
	}
	
	I2CSendByte(register_address);
	if(I2CError){
		I2CSendStop();
		return 0;
	}
	I2CSendStop();
	I2CSendStart();
	
	I2CSendByte((slave<<1)|I2C_READ_BIT);
	
	if(I2CError){
		I2CSendStop();
		return 0;
	}
	
	uint8_t data = I2CReceiveByte();
	I2CSend1();
	I2CSendStop();
	return data;
}

uint16_t I2C_read_word(uint8_t slave, uint8_t register_address){
	I2CSendStart();
	I2CSendByte(slave<<1|I2C_WRITE_BIT);
	
	if(I2CError){
		I2CSendStop();
		return 0;
	}
	
	I2CSendByte(register_address);
	if(I2CError){
		I2CSendStop();
		return 0;
	}
	I2CSendStop();
	
	I2CSendStart();
	
	I2CSendByte(slave<<1|I2C_READ_BIT);
	
	if(I2CError){
		I2CSendStop();
		return 0;
	}
	
	uint16_t data = I2CReceiveByte();
	I2CSend0();
	data = (data<<8) + I2CReceiveByte();
	I2CSend1();	
	I2CSendStop();
	return data;
}

void I2C_write_word(uint8_t slave, uint8_t register_address, uint16_t data){
	I2CSendStart();
	I2CSendByte(slave<<1|I2C_WRITE_BIT);
	
	if(I2CError){
		I2CSendStop();
		return;
	}
	
	I2CSendByte(register_address);
	if(I2CError){
		I2CSendStop();
		return;
	}
	
	I2CSendByte((data&0xff00)>>8);
	if(I2CError){
		I2CSendStop();
		return;
	}
	
	I2CSendByte(data&0xff);
	if(I2CError){
		I2CSendStop();
		return;
	}
	
	I2CSendStop();
	return;
}	
