/*
 * lcd_controller.c
 * 
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files 
 * (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */
 
 #include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "lcd_controller.h"

//to be shown on lcd.
uint8_t lcd_buffer[LCD_LINES*LCD_LINE_LENGTH+1];

uint8_t add_x = 0;
uint8_t add_y = 0;
uint8_t LCD_gXY = 1;

uint8_t buttons = 0;

//overflow interrupt: time to send uint8_t to LCD
ISR(TIMER0_OVF_vect){
	if (LCD_gXY == 1){
		LCD_gXY = 0;
		LCDGotoY(add_y);
		return;	
	}
	if(add_x==LCD_LINE_LENGTH){
		add_y++;
		if(add_y==LCD_LINES){
			add_y=0;
		}
		add_x = 0;
		LCDGotoY(add_y);
		return;	
	}

	//if((buttons&(1<<add_x/4))&&(add_x%4 == 1 || add_x%4 == 2))	
	if((add_y == 3)&&(buttons)&(1<<(5-add_x/4))&&(add_x%4==1||add_x%4==2))	
		LCDsendChar('#');		
	else 
		LCDsendChar(lcd_buffer[add_y*LCD_LINE_LENGTH+add_x]);
	add_x++;
	return;
}

void LCD_mark_buttons(uint8_t b){
	buttons = b;
}

void LCD_controller_init( void ){		
	LCDinit();
	LCDclr();	
	TIMSK0 = 0b1;
	TCCR0B = TIMER_PRESCALER_64;
	LCD_print_greeting();
}

void LCD_print_greeting(void){
	LCD_seek(0);	
	printf_P(PSTR("Press start to\n"));
	printf_P(PSTR("measure new data.\n"));
	printf_P(PSTR("Press prev/next to\n"));
	printf_P(PSTR("view measured data.\n"));
}


void clip(float * data, float boundary){
	if((*data) > boundary)
		*data = +INFINITY ;
	else if((*data) < - boundary)
		*data = -INFINITY ;
}

//prints data to stdout
void LCD_print_data(struct filtered_data * data){
	LCD_seek(0);
	
	float product = fabs(data->current * data->voltage);
	float ratio = fabs( data->voltage/data->current);
	if(fabs(data->current) < 0.0001)
		ratio = NAN;

	clip(&(data->voltage),19.999);
	clip(&(data->current),9.999);
	clip(&(data->voltage_der),0.9999);
	clip(&(data->current_der),0.9999);

	if(product > 29.99) 
		product = INFINITY ;
	if(ratio > 999.999) 
		ratio = INFINITY ;

	while(data->time>=10000*SECOND_TO_TIME){
		data->time = data->time %(10000*SECOND_TO_TIME);
	}

	printf_P(PSTR("n=%03u s=%02u t=%6.1fs\n"),data->set_index%1000, data->set%100, (float)data->time/SECOND_TO_TIME);
	if (fabs(data->voltage) < 1)
		printf_P(PSTR("V=% 6.1fmV  %+6.2f/s\n"),data->voltage*1000, data->voltage_der*1000);
	else
		printf_P(PSTR("V=% 6.0fmV  %+6.1f/s\n"),data->voltage*1000, data->voltage_der*1000);
	if (fabs(data->current) < 1)
		printf_P(PSTR("I=% 6.1fmA  %+6.2f/s\n"),data->current*1000, data->current_der*1000);
	else
		printf_P(PSTR("I=% 6.0fmA  %+6.1f/s\n"),data->current*1000, data->current_der*1000);
	
	if(product < 0.001)
		printf_P(PSTR("P=%6.0fuW R=%6.3f\n"),product * 1000 * 1000, ratio);
	else if(product < 0.01)
		printf_P(PSTR("P=%6.2fmW R=%6.3f\n"),product * 1000, ratio);
	else if(product < 0.1)
		printf_P(PSTR("P=%6.1fmW R=%6.3f\n"),product * 1000, ratio);
	else if(product < 1)
		printf_P(PSTR("P=%6.0fmW R=%6.3f\n"),product * 1000, ratio);
	else
		printf_P(PSTR("P=%6.2fW  R=%6.3f\n"),product, ratio);
}

uint8_t index_in = 0;
//buffer cursor to i.
void LCD_seek(uint8_t i){
	index_in = i;
	if(index_in >= LCD_LINE_LENGTH * LCD_LINES)
		index_in = 0;
}

//puts char to LCD
void LCD_putchar(char c){
	if(c == '\n' || c == '\0' || c == '\r'){
		for(;index_in % LCD_LINE_LENGTH;)
			lcd_buffer[index_in++] = ' ';
	} else {
		lcd_buffer[index_in++] = c;
		if(index_in >= LCD_LINE_LENGTH * LCD_LINES)
			index_in = 0;
	}
}
/*
void dec_exponent_form(float data, int8_t * exponent, int32_t * fraction){
	float abs_data = (data<0? -data:data);
	
	int8_t n = - 7;
	float c = 1E-7;
	for(; n < 7; c *= 10){
		if(abs_data < c)
			break;
		n++;
	}			
	*exponent = n;	
	*fraction = DEC_DIGITS * (data / c); 
	return;
}

void print_n_digits(uint8_t* buf, int32_t fraction, int8_t period, int8_t n){
	for(int8_t l = n; l>=0; l--){
		if(l == period)
			buf[l] = '.';
		else{
			buf[l] = '0' + fraction % 10;
			fraction /= 10;
		}
	}
}

void print_digits_and_prefix(uint8_t * buf, float data){
	#define EXP_OFFSET 11 
	int32_t fraction;
	int8_t exponent;
	dec_exponent_form(data,&exponent,&fraction);
	print_n_digits(buf,fabs(fraction),(exponent + EXP_OFFSET)%3 + 1,N_DIGITS);
	switch((exponent + EXP_OFFSET)/3){
		case 6: 
			buf[5] = 'M';
		break;
		case 5: 
			buf[5] = 'k';
		break;
		case 4: 
			buf[5] = ' ';
		break;
		case 3: 
			buf[5] = 'm';
		break;
		case 2: 
			buf[5] = 'u';
		break;
		case 1: 
			buf[5] = 'n';
		break;	
	};
	if(data<0)
		data = -data;
	int16_t i_data = (data * 10000);
	for(int8_t n = 4; n >= 0; n--){
		if(n == 3)
			buf[n] = '.';
		else{
			if(i_data == 0 && n < 2)
				buf[n] = ' ';
			else 
				buf[n] = '0' + i_data % 10;
			i_data /= 10;
		}		
	}
	buf[5] = 'm';
}

void print_index(uint8_t * buf, uint16_t index){
	buf[0] = ' ';
	buf[1] = 'n';
	buf[2] = ':';
	if(index>0){
		for(int8_t n = 9; n > 2; n--){
			if(index == 0)
				buf[n] = ' ';
			else 
				buf[n] = '0' + index % 10;
			index /= 10;
		}
	} else memcpy(&buf[3] , "current", 7);
}

void print_v_out(uint8_t * buf, int16_t v_out){
	buf[0] = ' ';
	buf[1] = 'V';
	buf[2] = ':';
	if(v_out < 0){
		buf[3] = '-';
		v_out = -v_out;
	} else buf[3] = '+';
	for(int8_t n = 9; n > 3; n--){
		if(v_out == 0)
			buf[n] = ' ';
		else 
			buf[n] = '0' + v_out % 10;
		v_out /= 10;
	}
}

void print_time(uint8_t * buf, uint32_t time){
	time = 2560*time/F_CPU;
	for(int8_t n = 8; n >= 2; n--){
		if(n == 7)
			buf[n] = '.';
		else{
			if(time == 0 && n < 6)
				buf[n] = ' ';
			else 
				buf[n] = '0' + time % 10;
			time /= 10;
		}		
	}
	buf[0] = ' ';
	buf[1] = 't';
	buf[2] = ':';
	buf[9] = 's';
}

void print_current(uint8_t * buf, float current){
	buf[0] = ' ';
	buf[1] = 'I';
	buf[2] = ':';
	print_digits_and_prefix(&buf[3],current);
	buf[9] = 'A';
}

void print_voltage(uint8_t * buf, float voltage){
	buf[0] = ' ';
	buf[1] = 'U';
	buf[2] = ':';
	print_digits_and_prefix(&buf[3],voltage);
	buf[9] = 'V';
}

void print_power(uint8_t * buf, float power){
	buf[0] = ' ';
	buf[1] = 'P';
	buf[2] = ':';
	print_digits_and_prefix(&buf[3],power);
	buf[9] = 'W';
}

void print_current_der(uint8_t * buf, float current_der){
	buf[0] = 'd';
	buf[1] = 'I';
	buf[2] = ':';
	print_digits_and_prefix(&buf[3],current_der);
	buf[9] = 'A';
}

void print_voltage_der(uint8_t * buf, float voltage_der){
	buf[0] = 'd';
	buf[1] = 'U';
	buf[2] = ':';
	print_digits_and_prefix(&buf[3],voltage_der);
	buf[9] = 'V';
}

void print_power_der(uint8_t * buf, float power_der){
	buf[0] = 'd';
	buf[1] = 'P';
	buf[2] = ':';
	print_digits_and_prefix(&buf[3],power_der);
	buf[9] = 'W';
}

void LCD_controller_print_data(struct filtered_data * data, int8_t page, uint16_t index, int8_t upper ){
	
	int8_t u = 0;//	int8_t u = (upper%2)*2;
	print_index(&lcd_buffer[u][0],index);
	print_time(&lcd_buffer[u][10],data->time);
	print_voltage_der(&lcd_buffer[u+2][10],data->voltage_der);
	print_voltage(&lcd_buffer[u+2][0],data->voltage);
	print_v_out(&lcd_buffer[u+1][10],data->v_out_meas);
	print_v_out(&lcd_buffer[u+1][0],data->v_in_meas);
	
//	switch (page){
		case 2:
			print_power(&lcd_buffer[u+1][0],data->voltage * data->current);
			print_power_der(&lcd_buffer[u+1][10],data->current_der * data->voltage + data->current * data->voltage_der);
		break;
		case 3:
			print_voltage_der(&lcd_buffer[u+1][0],data->voltage_der);
			print_current_der(&lcd_buffer[u+1][10],data->current_der);
		break;
//		default:
			//print_voltage(&lcd_buffer[u+1][0],data->voltage);
//			if(u){
//			}
						
//		break;
//	}	
//}
*/
