/*
 * log.c
 * 
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files 
 * (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */
 
#include "log.h"
#include "spi.h"
#include <stdio.h>
#include <avr/pgmspace.h>

#define MAX_ADDRESS 0x10000
#define SECTOR_SIZE 0x01000
#define DATA_SIZE 0x20

uint32_t cur_abs_index = 0;
uint16_t cur_set_index = 0;
uint16_t cur_set = 0;

//read flash for previos set
void log_init(void){
	cur_abs_index = 0;
	cur_set_index = 0;
	cur_set = 0;

	struct filtered_data data;
	for(uint32_t i = 0 ; i < MAX_ADDRESS; i+= SECTOR_SIZE){
		spiflash_readblock((uint8_t * )&data,i,sizeof(struct filtered_data));

		if(data.abs_index == 0xFFFFFFFF)
			continue;

		if(cur_abs_index < data.abs_index){
			cur_abs_index = data.abs_index;
			cur_set_index = data.set_index;
			cur_set = data.set;
		}

		for(uint32_t j = SECTOR_SIZE/4, k = i + SECTOR_SIZE/2; j >= DATA_SIZE/2; j/=2){
			spiflash_readblock((uint8_t * )&data,k,sizeof(struct filtered_data));

			if(data.abs_index == 0xFFFFFFFF){
				k-=j;
				continue;
			}
			
			if(cur_abs_index < data.abs_index){
				cur_abs_index = data.abs_index;
				cur_set_index = data.set_index;
				cur_set = data.set;
			}
			
			k+=j;
		}
	}
}

void log_log_data(struct filtered_data * data){
	data->abs_index = ++cur_abs_index;
	data->set_index = ++cur_set_index;
	data->set = cur_set;

	uint32_t address = ((cur_abs_index-1) * DATA_SIZE) % MAX_ADDRESS;
	if(address % SECTOR_SIZE == 0)
		spiflash_sector_erase(address);
		
	spiflash_writeblock((uint8_t *)data, address, 32);
}

//get logged data at abs_index, return false if index out of bounds
int8_t log_get_data(struct filtered_data * data, uint32_t abs_index){
	if(abs_index == 0)
		return 0;
	uint32_t address = ((abs_index-1) * DATA_SIZE) % MAX_ADDRESS;
	spiflash_readblock((uint8_t * )data,address,sizeof(struct filtered_data));
	if(data->abs_index != abs_index)
		return 0;
	return 1;
}

/*
//get logged data at set_index of set, return false if index out of bounds
int8_t log_get_data(struct filtered_data * data, uint16_t set, uint16_t set_index){


}*/

//get logged data at beginning or end of set index set, return false if index out of bounds
int8_t log_get_data_set_boundary(struct filtered_data * data, uint16_t set, uint8_t first){

	uint32_t abs_index = 0;
	uint16_t set_index = 0;
	for(uint32_t i = 0 ; i < MAX_ADDRESS; i+= SECTOR_SIZE){
		spiflash_readblock((uint8_t * )data,i,sizeof(struct filtered_data));

		if(data->abs_index == 0xFFFFFFFF)
			continue;

		if(set == data->set){
			if(first && (set_index > data->set_index || set_index == 0)){
				abs_index = data->abs_index;
				set_index = data->set_index;			}
			if(first == 0 && set_index < data->set_index){
				abs_index = data->abs_index;
				set_index = data->set_index;
			}
		}

		for(uint32_t j = SECTOR_SIZE/4, k = i + SECTOR_SIZE/2; j >= DATA_SIZE/2; j/=2){
			spiflash_readblock((uint8_t * )data,k,sizeof(struct filtered_data));

			if(data->abs_index == 0xFFFFFFFF){
				k-=j;
				continue;
			}
			
			if(first){
				if(data->set >= set){
					if(data->set == set){
						if(data->set_index < set_index || set_index == 0){
								abs_index = data->abs_index;
								set_index = data->set_index;
						}
					}					
					k-=j;
				} else
					k+=j;
			} else {
				if(data->set > set)					
					k-=j;
				else {
					if(data->set == set){
						if(data->set_index > set_index){
								abs_index = data->abs_index;
								set_index = data->set_index;
						}
					}	
					k+=j;
				}			
			}
		}
	}
	if(set_index == 0)
		return 0;
	return log_get_data(data,abs_index);
}
