/*
 * main.c
 * 
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files 
 * (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */
 
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "avr/wdt.h"
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>

#include "usb_serial.h"
#include "adc_controller.h"
#include "spi.h"
#include "layout.h"
#include "log.h"
#include "lcd_controller.h"
#include "touch_controller.h"

#define CPU_PRESCALE(n)	(CLKPR = 0x80, CLKPR = (n))


#define OUTPUT_TO_USB 0x1
#define OUTPUT_TO_LCD 0x2

uint8_t stream_output = OUTPUT_TO_USB | OUTPUT_TO_LCD;
static int stream_putchar(char c, FILE *stream);
static int stream_getchar(FILE *stream);

static FILE usb_stream = FDEV_SETUP_STREAM(stream_putchar, stream_getchar, _FDEV_SETUP_RW);

static int	stream_putchar(char c, FILE *stream){
	if(stream_output & OUTPUT_TO_USB)
		usb_serial_putchar(c);
	if(stream_output & OUTPUT_TO_LCD)
		LCD_putchar(c);
	return 0;
}

static int stream_getchar(FILE *stream){
	int n = usb_serial_getchar();
	while(n == -1)
		n = usb_serial_getchar();
	return n;
}

enum process{
	GREETING,
	MEASURING,
	BROWSING,
	TESTING,
	CHARGING
};

struct state{
	enum process current_process;
	uint8_t delay;
	uint32_t abs_index;
	uint16_t set_index;
	uint16_t set;
	
	uint8_t code_index;
} current_state;

uint8_t const code[] PROGMEM =  {'s','s','s','t','+','-','t','t','r'};
int16_t const test_points_voltage[N_TEST_POINTS] PROGMEM = {20,30,40,50,60,70,80,90,100,110,
															120,130,140,150,160,170,180,190,200,210,
															220,230,235,
															255,260,270,280,290,300,310,320,330,340,
															350,360,370,380,390,400,410,420,430,440}; // mV

void print_info(float bat){
	uint16_t buf = eeprom_read_word(&bootloader_version);
	LCD_seek(0);
	printf_P(PSTR(" bootloader v %u\n"),buf);
	printf_P(PSTR(" firmware   v %u\n"),/*FIRMWARE_REVISION_TAG*/155);

	printf_P(PSTR(" battery %fV\n"),(double)bat);
	buf = eeprom_read_byte(&serial_number);
	printf_P(PSTR(" serial number %u\n"),buf);
}

void start_new_set(void){
	if(cur_set_index || cur_set == 0){
		cur_set ++;
		cur_set_index = 0;
	}
	
	current_state.abs_index = cur_abs_index + 1;
	current_state.set_index = 1;
	current_state.set = cur_set;

	reset_timer();
}

void task(uint8_t cmd){
	uint8_t stream_output_buf = stream_output;

	if(pgm_read_byte(&code[current_state.code_index]) == cmd && cmd>=32){
		current_state.code_index ++;
		stream_output = stream_output_buf;
		
		if(current_state.code_index >= sizeof(code)){
			current_state.code_index = 0;
			cmd = 'T';
		}
	} else 
		current_state.code_index = 0;

	switch(cmd){
		case 'h':
			stream_output = OUTPUT_TO_USB;
			
			printf_P(PSTR("commands:\n"));
			printf_P(PSTR("\th - help\n"));
			
			printf_P(PSTR("commands that duplicate buttons\n"));
			printf_P(PSTR("\tr - start measurements\n"));
			printf_P(PSTR("\ts - stop measurements\n"));
			printf_P(PSTR("\t- - previous measurement\n"));
			printf_P(PSTR("\t+ - next measurement\n"));
			printf_P(PSTR("\tp - previous set\n"));
			printf_P(PSTR("\tn - next set\n"));
			printf_P(PSTR("\tt - trigger - take sample\n"));
			printf_P(PSTR("\to - turn off\n"));

			printf_P(PSTR("usb commands\n"));
			printf_P(PSTR("\ti - version and status\n"));
			printf_P(PSTR("\tb - jump to bootloader\n"));
			//printf_P(PSTR("\te - turn on/off display to usb echo\n"));
			printf_P(PSTR("\tR - read calibration from device\n"));
			printf_P(PSTR("\tW - write calibration to device\n"));
			printf_P(PSTR("\tT - experimental setup test\n"));

			printf_P(PSTR("flash commands\n"));
			printf_P(PSTR("\tz - flash erase\n"));
			printf_P(PSTR("\tl - flash read\n"));
			printf_P(PSTR("\tP - flash print saved samples\n"));
			
			printf_P(PSTR("debug commands\n"));
			printf_P(PSTR("\tm - print ram contents\n"));
			
			stream_output = stream_output_buf;
		break;
		case 'm':
			printf("%x\n",SP);
			for(uint16_t i = 0x100; i<=0x4FF;i++){
				if(i%32==0)		
					usb_serial_putchar('\n');
				printf("%02x",*(uint8_t *)i);
			}
		break;
		case 's':
			stream_output = OUTPUT_TO_USB;
			printf_P(PSTR("measurements done\n"));
			stream_output = stream_output_buf;
			current_state.current_process = GREETING;
			LCD_print_greeting();
		break;
		case 'r':		
			if(current_state.current_process != MEASURING){

				stream_output = OUTPUT_TO_USB;
				printf_P(PSTR("measurements starting\n"));
				stream_output = stream_output_buf;
				current_state.current_process = MEASURING;
				
				start_new_set();
				
			}
		break;
		case '-':
			switch(current_state.current_process){
				case GREETING:
					if(cur_set == 0)
						break;
					
					current_state.abs_index = cur_abs_index;
					current_state.set_index = cur_set_index;
					current_state.set = cur_set;
					if(log_get_data(&latest_data,cur_abs_index)==0)
						break;
						
					current_state.current_process = BROWSING;					
					LCD_print_data(&latest_data);					
				break;
				case MEASURING:
					if(current_state.abs_index <= 1)
						break;
						
					if(log_get_data(&latest_data,current_state.abs_index - 1))
						if(current_state.set == latest_data.set){
							current_state.set_index = latest_data.set_index;
							current_state.abs_index --;
							LCD_print_data(&latest_data);					
						}					
				break;
				case BROWSING:
					if(current_state.abs_index <= 1)
						break;
					
					if(log_get_data(&latest_data,current_state.abs_index - 1)){
						current_state.set = latest_data.set;
						current_state.set_index = latest_data.set_index;
						current_state.abs_index --;
						LCD_print_data(&latest_data);					
					}					
				break;
				default:
				break;
			}
		break;
		case '+':
			switch(current_state.current_process){
				case MEASURING:
					if(current_state.abs_index > cur_abs_index)
						break;
					
					if(log_get_data(&latest_data,current_state.abs_index + 1)){
						if(current_state.set == latest_data.set){
							current_state.set_index = latest_data.set_index;
							current_state.abs_index ++;
							LCD_print_data(&latest_data);					
						} 
					}else{
						current_state.set_index = cur_set_index + 1;
						current_state.abs_index = cur_abs_index + 1;
					}
				break;
				case BROWSING:
					if(current_state.abs_index > cur_abs_index)
						break;
					
					if(log_get_data(&latest_data,current_state.abs_index + 1)){
						current_state.set = latest_data.set;
						current_state.set_index = latest_data.set_index;
						current_state.abs_index ++;
						LCD_print_data(&latest_data);
						break;	
					} 
					LCD_print_greeting();
					current_state.current_process = GREETING;
				break;
				default:
				break;
			}
		break;
		case 'p':
			if(current_state.current_process == BROWSING){
				if(current_state.abs_index <= 1)
					break;
				
				if(current_state.set > 1){
					if(log_get_data_set_boundary(&latest_data,current_state.set-1,0)){
						current_state.set = latest_data.set;
						current_state.set_index = latest_data.set_index;
						current_state.abs_index = latest_data.abs_index;
						LCD_print_data(&latest_data);
						break;
					} else if (log_get_data_set_boundary(&latest_data,current_state.set,1)){
						current_state.set = latest_data.set;
						current_state.set_index = latest_data.set_index;
						current_state.abs_index = latest_data.abs_index;
						LCD_print_data(&latest_data);
						break;
					}
				} else if (log_get_data_set_boundary(&latest_data,current_state.set,1)){
					current_state.set = latest_data.set;
					current_state.set_index = latest_data.set_index;
					current_state.abs_index = latest_data.abs_index;
					LCD_print_data(&latest_data);
					break;
				}
			}
			if(current_state.current_process == MEASURING){
				if (log_get_data_set_boundary(&latest_data,current_state.set,1)){
					current_state.set = latest_data.set;
					current_state.set_index = latest_data.set_index;
					current_state.abs_index = latest_data.abs_index;
					LCD_print_data(&latest_data);
					break;
				}
 			}
		break;
		case 'n':
			if(current_state.current_process == BROWSING){
				if(current_state.abs_index > cur_abs_index)
					break;
				
				if(current_state.set < cur_set){
					if(log_get_data_set_boundary(&latest_data,current_state.set+1,1)){
						current_state.set = latest_data.set;
						current_state.set_index = latest_data.set_index;
						current_state.abs_index = latest_data.abs_index;
						LCD_print_data(&latest_data);
						break;
					} 
				} else if (log_get_data_set_boundary(&latest_data,current_state.set,0)){
					current_state.set = latest_data.set;
					current_state.set_index = latest_data.set_index;
					current_state.abs_index = latest_data.abs_index;
					LCD_print_data(&latest_data);
					break;
				}
			}
			
			if(current_state.current_process == MEASURING){
				current_state.abs_index = cur_abs_index + 1;
				current_state.set_index = cur_set_index + 1; 
 			}
		break;
		case 'i':
			stream_output = OUTPUT_TO_USB;
			printf_P(PSTR("version and status:\n"));
			float v_bat = adc_measure(ADC_CHANNEL_V_BAT,1)*2;
		 	print_info(v_bat);
			stream_output = stream_output_buf;			

		break;
		case 'b':
			eeprom_write_byte(&stay_in_bootloader,1);
			_delay_ms(100);
			boot();
		break;
		case 't':
			if(current_state.current_process == MEASURING && current_state.delay == 0){
				current_state.delay = 1;
				log_log_data(&latest_data);
				current_state.abs_index = cur_abs_index + 1;
				current_state.set_index = cur_set_index + 1;
				latest_data.set_index++;
				LCD_print_data(&latest_data);
			}
		break;
		case 'f':
			printf_P(PSTR("flash status %x\n"),spiflash_readstatusreg());
		break;
		case 'd':
			printf_P(PSTR("flash id %x\n"),spiflash_read_id());
		break;
		case 'z':
			printf_P(PSTR("erasing flash ..."));
			spiflash_chip_erase();
			_delay_ms(1);
			log_init();
			printf_P(PSTR(" done \n"));
		break;
		case 'l':
			stream_output = OUTPUT_TO_USB;
			printf_P(PSTR("reading flash ...\n"));
			for (uint32_t n = 0; n < 2024;n++){
				wdt_reset();
				uint8_t buf[32];
				spiflash_readblock(buf,32*n,32);
				for(uint8_t i = 0; i < 32; i++)
					printf_P(PSTR("%02x"),buf[i]);
				printf("\n");
			}
			stream_output = stream_output_buf;
		break;
		case 'o':
			LCD_seek(0);
			printf_P(PSTR("my mind is going ...\n \n \n \n"));
			eeprom_write_byte(&stay_in_bootloader,0);
			//usb_serial_putchar_nowait('\n');
			_delay_ms(100);
			POWER_SW(0);
			_delay_ms(100);
		break;
		case 'c':
			current_state.current_process = CHARGING;
		break;
		case 'R':
			stream_output = OUTPUT_TO_USB;
			printf_P(PSTR("Reading calibration from eeprom\n"));
			
			for(uint8_t channel = 0; channel < 7; channel++){
				for(uint8_t attn = 0; attn < 2; attn++){
					for(uint8_t range = 0; range < 6; range++){
						printf_P(PSTR("%d %d"),
						(int16_t)eeprom_read_word((uint16_t *)&(v_calibration[channel][attn][range][0])),
						(int16_t)eeprom_read_word((uint16_t *)&(v_calibration[channel][attn][range][1])));
						usb_serial_putchar('\t');
					}				
					usb_serial_putchar('\n');
				}
				usb_serial_putchar('\n');
			}
			
			printf_P(PSTR("%f\n\n"),eeprom_read_float(&i_resistor_calibration));
		
			
			for(uint8_t i = 0; i < N_TEST_POINTS; i++){	
				printf_P(PSTR("%d\t"),(int16_t)eeprom_read_word((uint16_t *)&test_calibration[i]));
			}
			
			printf_P(PSTR("\n\n%d\n"),(int16_t)eeprom_read_word((uint16_t *)&test_end_voltage));

			stream_output = stream_output_buf;		
		break;
		case 'W':;
			int16_t buf;
			
			for(uint8_t channel = 0; channel < 7; channel++)
				for(uint8_t attn = 0; attn < 2; attn++)
					for(uint8_t range = 0; range < 6; range++){
						wdt_reset();
						scanf("%d",&buf);
						eeprom_write_word((uint16_t *)&(v_calibration[channel][attn][range][0]),(uint16_t)buf);
						scanf("%d",&buf);
						eeprom_write_word((uint16_t *)&(v_calibration[channel][attn][range][1]),(uint16_t)buf);
					}				
			
			float fbuf;
			scanf("%f",&fbuf);
			eeprom_write_float(&i_resistor_calibration,fbuf);
		
			for(uint8_t i = 0; i < N_TEST_POINTS; i++){	
				wdt_reset();
				scanf("%d",&buf);
				eeprom_write_word((uint16_t *)&test_calibration[i],(uint16_t)buf);
			}
		
			scanf("%d",&buf);
			eeprom_write_word((uint16_t *)&test_end_voltage,(uint16_t)buf);
	
		break;
		case 'P':
			stream_output = OUTPUT_TO_USB;
			for(uint32_t i = cur_abs_index; i > 0; i--){
				wdt_reset();
				if(log_get_data(&latest_data,i))
					LCD_print_data(&latest_data);
			}
			stream_output = stream_output_buf;
		break;
		case 'T':
			current_state.current_process = TESTING;
			
			start_new_set();
		break;
	}

}

void my_init(void) __attribute__((naked)) __attribute__ ((section (".init1")));
void my_init(void) {
   // Break into the C startup so I can clear SRAM to
   // known values making it easier to see how it is used
   uint8_t *p;
   uint16_t i;
   
   for (i=0x100; i < RAMEND; i++) {
      p = (uint8_t *)i;
      *p = 0x5A;
   }
} 


int main(void)
{
	// set for 8 MHz clock
	CPU_PRESCALE(0);
	
	MCUCR = (1 << IVCE) | (0 << PUD);
	MCUCR = (0 << IVSEL) | (0 << IVCE) | (0 << PUD);
	
	initIO();
	usb_init();
	timer_init();
	setup_spi(SPI_MODE_0, SPI_MSB, SPI_NO_INTERRUPT, SPI_MSTR_CLK2);
	spiflash_writestatusreg(0); //remove write protection
	touch_controller_init();
	sei();
	
	_delay_ms(500); //wait for usb

	stdout = &usb_stream; 
	stdin = &usb_stream;
	
	LCD_controller_init();
	log_init();
	
	if(usb_configured())
		current_state.current_process = CHARGING;
	else	
		current_state.current_process = GREETING;
	
	
	while (1) {
		wdt_reset();
		
		if(!POWER_SENSE()){
			task('o');
			continue;
		}
		
		int n = usb_serial_getchar();
		if (n >= 0)
			task(n);
		
		static uint8_t last_key = 0;
		static uint32_t last_key_time = 0;
		
		if(TOUCH_INT()){
			uint8_t x = get_key_status();
			//printf_P(PSTR("key pressed %x \n"),x);
			LCD_mark_buttons(x);
			last_key = 0; 
			last_key_time = time_now();
			switch(x){
				case KEY_START:	task('r');	break;
				case KEY_STOP:	task('s');	break;
				case KEY_PREV:	task('-'); last_key = 'p';	break;
				case KEY_NEXT:	task('+'); last_key = 'n';	break;
				case KEY_TRIG:	task('t');  break;
				case KEY_NONE:	task(' ');  break;
				case KEY_GUARD:	task(' '); 	break;
			}
		} else if (last_key ) {
			if((time_now() - last_key_time) > 32768){
				last_key_time += 32768;
				task(last_key);
			}
		}	
	
		if(current_state.current_process == TESTING){
			filter_take_new_data();

			static int16_t latest_voltage = 0; //in mV-s
 			static float greatest_voltage = 0; //in V-s
			static uint8_t i_test_points = 0; 
			static uint8_t n_nan_test_points = 0;
			static uint8_t n_failed_test_points = 0;
			static uint8_t n_failed_current_points = 0;
						 
			if(latest_data.voltage - latest_voltage/1000.F >= 0.005){
				latest_voltage += 5;
				log_log_data(&latest_data);
			} else if(latest_voltage/1000.F - latest_data.voltage >= 0.005){
				latest_voltage -= 5;
				log_log_data(&latest_data);
			}
			
			if(i_test_points<N_TEST_POINTS && latest_voltage >= (int16_t)pgm_read_word(&test_points_voltage[i_test_points])){
				if(isnan(latest_data.voltage_der))
					n_nan_test_points++;
				else{
					float calibration = ((int16_t)eeprom_read_word((uint16_t *)&test_calibration[i_test_points]));
					if(fabs((calibration - latest_data.voltage_der * 1000000)) > 200)
						n_failed_test_points++;
				}
				if(latest_data.current<0.0057||latest_data.current>0.0063)
					n_failed_current_points++;
				i_test_points++;
			}
			
			if(latest_data.voltage > greatest_voltage)
				greatest_voltage = latest_data.voltage;
			
			LCD_seek(0);

			uint16_t buf = eeprom_read_word(&serial_number);

			printf_P(PSTR("serial number %u\n"),buf);
			
			float v_bat = adc_measure(ADC_CHANNEL_V_BAT,1)*2;			
			
			if(v_bat>3.6)
				printf_P(PSTR("battery %.2fV ok\n"),(double)v_bat);
			else
				printf_P(PSTR("battery %.2fV low\n"),(double)v_bat);
			
			
			if(latest_data.current>0.0057&&latest_data.current<0.0063&&n_failed_current_points<=2)
				printf_P(PSTR("V %.1f current ok\n"),latest_data.voltage*1000);
			else
				printf_P(PSTR("V %.1f current fail\n"),latest_data.voltage*1000);
		
			if(i_test_points >= N_TEST_POINTS && 
				fabs(greatest_voltage - 0.001F*((int16_t)eeprom_read_word((uint16_t *)&test_end_voltage)))<0.02 ){
					if(n_nan_test_points + n_failed_test_points > 4)
						printf_P(PSTR("test done fail:%u/%u\n"),n_nan_test_points + n_failed_test_points, N_TEST_POINTS);
					else
						printf_P(PSTR("test done ok:  %u/%u\n"),N_TEST_POINTS - n_nan_test_points - n_failed_test_points, N_TEST_POINTS);
			} else {
				printf_P(PSTR("testing %u/%u ok %u\n"),i_test_points,N_TEST_POINTS,i_test_points - n_nan_test_points - n_failed_test_points);
			}
			
			


/*			static uint8_t delay = 0;

			if(((delay++)/16)%2){

				uint16_t buf = eeprom_read_word(&bootloader_version);


			} else {

				if(latest_data.voltage > 0.9999)
					latest_data.voltage = 0.9999;
	
				if(latest_data.voltage_der > 0.009999)
					latest_data.voltage_der = 0.009999;

				if(latest_data.voltage < -0.9999)
					latest_data.voltage = -0.9999;
	
				if(latest_data.voltage_der < -0.009999)
					latest_data.voltage_der = -0.009999;

				LCD_seek(0);
				if(latest_data.current > 0.006 && latest_data.current < 0.007)	
					printf_P(PSTR("current source ok\n"));
				else 
					printf_P(PSTR("current source fail\n"));
		
				printf_P(PSTR("U=% 6.1fmV  %+3.3f/s\n"),latest_data.voltage*1000, latest_data.voltage_der*1000);
				printf_P(PSTR("NLE stable\n"));
				if(latest_data.voltage > 0.5 && latest_data.voltage < 0.59 && latest_data.current > 0.0058 && latest_data.current < 0.007)	
					printf_P(PSTR("testing done: ok\n"));
				else 
					printf_P(PSTR("testing in progress\n"));
			}		
*/
		}

		if(current_state.current_process == CHARGING){
			float v_bat = adc_measure(ADC_CHANNEL_V_BAT,1)*2;
		 	uint8_t stream_output_buf = stream_output;
			stream_output = OUTPUT_TO_LCD;
		 	print_info(v_bat);
		 	stream_output = stream_output_buf;
		}
		if(current_state.current_process == MEASURING){
			uint8_t stream_output_buf = stream_output;
			stream_output = OUTPUT_TO_USB;
			
			volatile uint32_t time = time_now();
			uint32_t delay = time - time%6000 + 6000;
			while(time_now() < delay)
				_delay_us(1);
			
			
			filter_take_new_data();
			stream_output = stream_output_buf;
						
			if(current_state.delay){
				current_state.delay --;
				continue;
			}
			
			latest_data.abs_index = cur_abs_index+1;
			latest_data.set_index = cur_set_index+1;
			latest_data.set = cur_set;
			
			if(latest_data.abs_index == current_state.abs_index)
				LCD_print_data(&latest_data);
		}
		
	}
}
		
