/* 
 * spi.c
 * 
 * Copyright (c) 2009 Andrew Smallbone <andrew@rocketnumbernine.com>
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sub-license, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <util/delay.h>
#include "spi.h"
#include "lcd_lib.h"
#include "layout.h"

#ifdef __cplusplus
extern "C"{
#endif

#ifdef __ARDUINO__
#include <wiring.h>
#endif


void setup_spi(uint8_t mode, int dord, int interrupt, uint8_t clock)
{
  // specify pin directions for SPI pins on port B
  if (clock == SPI_SLAVE) { // if slave SS and SCK is input
    DDRB &= ~(1<<SPI_MOSI_PIN); // input
    DDRB |= (1<<SPI_MISO_PIN); // output
    DDRB &= ~(1<<SPI_SS_PIN); // input
    DDRB &= ~(1<<SPI_SCK_PIN);// input
  } else {
    DDRB |= (1<<SPI_MOSI_PIN); // output
    DDRB &= ~(1<<SPI_MISO_PIN); // input
	PORTB |= (1<<SPI_MISO_PIN); // enable pull-up
    DDRB |= (1<<SPI_SCK_PIN);// output
    DDRB |= (1<<SPI_SS_PIN);// output
  }
  SPCR = ((interrupt ? 1 : 0)<<SPIE) // interrupt enabled
    | (1<<SPE) // enable SPI
    | (dord<<DORD) // LSB or MSB
    | (((clock != SPI_SLAVE) ? 1 : 0) <<MSTR) // Slave or Master
    | (((mode & 0x02) == 2) << CPOL) // clock timing mode CPOL
    | (((mode & 0x01)) << CPHA) // clock timing mode CPHA
    | (((clock & 0x02) == 2) << SPR1) // cpu clock divisor SPR1
    | ((clock & 0x01) << SPR0); // cpu clock divisor SPR0
  SPSR = (((clock & 0x04) == 4) << SPI2X); // clock divisor SPI2X
}

void disable_spi()
{
  SPCR = 0;
}


// transceive data over SPI
// sends /out/ and returns received data
uint8_t trans_spi(uint8_t out)
{
  SPDR = out;
  while (!(SPSR & (1<<SPIF)));
  return SPDR;
}


uint8_t spiflash_readstatusreg(void) {
	spi_CS_en();
	trans_spi(SPIFLASH_READSTATUSREG);
	uint8_t out=trans_spi(SPIFLASH_READSTATUSREG);
	spi_CS_dis();
	return out;
}

// enable writing
void spiflash_enwrite(void) {
	spi_CS_en();
	trans_spi(SPIFLASH_ENWRITE);
	spi_CS_dis();
}

// disable writing
void spiflash_diswrite(void) {
	spi_CS_en();
	trans_spi(SPIFLASH_DISWRITE);
	spi_CS_dis();
}

void spiflash_writestatusreg(uint8_t data) {
	spi_CS_en();
	trans_spi(SPIFLASH_ENWRITESTATUSREG);
	spi_CS_dis();
	_delay_us(10);
	spi_CS_en();
	trans_spi(SPIFLASH_WRITESTATUSREG);
	trans_spi(data);
	spi_CS_dis();
}

// reads to /data/ from SPI flash /address/, /count/ bytes
void spiflash_readblock(uint8_t *data, uint32_t address, uint8_t count) {
	spi_CS_en();
	trans_spi(SPIFLASH_READ);
	trans_spi((address&0xff0000)>>16);
	trans_spi((address&0xff00)>>8);
	trans_spi((address&0xff));
	for (uint8_t i=0; i<count; i++) {
		data[i]=trans_spi(0);
	}
	spi_CS_dis();
}

// writes /data/ of size /count/ to SPI flash /address/
void spiflash_writeblock(uint8_t *data, uint32_t address, uint8_t count) {
	spiflash_enwrite();
	spi_CS_en();
	trans_spi(SPIFLASH_AUTOINCR);
	trans_spi((address&0xff0000)>>16);
	trans_spi((address&0xff00)>>8);
	trans_spi((address&0xff));
	trans_spi(data[0]);
	spi_CS_dis();
	while (spiflash_isbusy());		

	for (uint8_t i=1; i<count; i++) {
		spi_CS_en();
		trans_spi(SPIFLASH_AUTOINCR);	
		trans_spi(data[i]);
		spi_CS_dis();
		while (spiflash_isbusy());		

	}
	spiflash_diswrite();
	while (spiflash_isbusy());
}

void spiflash_writebyte(uint32_t address, uint8_t data) {
	spiflash_enwrite();
	spi_CS_en();
	trans_spi(SPIFLASH_BYTEPROG);
	trans_spi((address&0xff0000)>>16);
	trans_spi((address&0xff00)>>8);
	trans_spi((address&0xff));
	trans_spi(data);
	spi_CS_dis();
	while (spiflash_isbusy());		
	spiflash_diswrite();
}

// read flash ID
uint16_t spiflash_read_id(void)
{
	spi_CS_en();
	trans_spi(SPIFLASH_READID);
	trans_spi(0x00);
	trans_spi(0x00);
	trans_spi(1);
	uint16_t out=(trans_spi(0)<<8)+(trans_spi(0)&0xff);
	spi_CS_dis();
	return out;
}

//erase flash sector (4 KiB) at /address/
void spiflash_sector_erase(uint32_t address){
	spiflash_enwrite();
	spi_CS_en();
	trans_spi(SPIFLASH_SECTORERASE);
	trans_spi((address&0xff0000)>>16);
	trans_spi((address&0xf000)>>8);
	trans_spi((0x00));
	spi_CS_dis();
	while (spiflash_isbusy()) {
		_delay_us(200);
	spiflash_diswrite();
	}
}

// erase flash block (32 KiB) at address
void spiflash_block_erase(uint32_t address){
	spiflash_enwrite();
	spi_CS_en();
	trans_spi(SPIFLASH_BLOCKERASE);
	trans_spi((address&0xff0000)>>16);
	trans_spi((address&0x8000)>>8);
	trans_spi(0);
	spi_CS_dis();
	while (spiflash_isbusy()) {
		_delay_us(200);
	}
	spiflash_diswrite();
}

void spiflash_chip_erase(void){
	spiflash_enwrite();
	spi_CS_en();
	trans_spi(SPIFLASH_CHIPERASE);
	spi_CS_dis();
	while (spiflash_isbusy()) {
		_delay_us(500);
	}
	spiflash_diswrite();
}
