/* 
 * Copyright (c) 2009 Andrew Smallbone <andrew@rocketnumbernine.com>
 * Copyright (c) 2012 Hedgehog LLC, www.hedghog.ee/ipho
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sub-license, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#ifndef _spi_h__
#define _spi_h__

#include <avr/io.h>

#ifdef __cplusplus
extern "C"{
#endif

// SPI clock modes
#define SPI_MODE_0 0x00 /* Sample (Rising) Setup (Falling) CPOL=0, CPHA=0 */
#define SPI_MODE_1 0x01 /* Setup (Rising) Sample (Falling) CPOL=0, CPHA=1 */
#define SPI_MODE_2 0x02 /* Sample (Falling) Setup (Rising) CPOL=1, CPHA=0 */
#define SPI_MODE_3 0x03 /* Setup (Falling) Sample (Rising) CPOL=1, CPHA=1 */

// data direction
#define SPI_LSB 1 /* send least significant bit (bit 0) first */
#define SPI_MSB 0 /* send most significant bit (bit 7) first */

// whether to raise interrupt when data received (SPIF bit received)
#define SPI_NO_INTERRUPT 0
#define SPI_INTERRUPT 1

// slave or master with clock divisor
#define SPI_SLAVE 0xF0
#define SPI_MSTR_CLK4 0x00 /* chip clock/4 */
#define SPI_MSTR_CLK16 0x01 /* chip clock/16 */
#define SPI_MSTR_CLK64 0x02 /* chip clock/64 */
#define SPI_MSTR_CLK128 0x03 /* chip clock/128 */
#define SPI_MSTR_CLK2 0x04 /* chip clock/2 */
#define SPI_MSTR_CLK8 0x05 /* chip clock/8 */
#define SPI_MSTR_CLK32 0x06 /* chip clock/32 */

// SPI flash commands
#define SPIFLASH_READ	0x03 // read
#define SPIFLASH_HSREAD	0x0B // high-speed read
#define SPIFLASH_SECTORERASE	0x20 // sector (4KiB) erase
#define SPIFLASH_BLOCKERASE	0x52 // block (32 KiB) erase
#define SPIFLASH_CHIPERASE	0x60 // chip erase
#define SPIFLASH_BYTEPROG	0x02 // byte program
#define SPIFLASH_AUTOINCR	0xAF // auto address increment
#define SPIFLASH_READSTATUSREG	0x05 // read status register
#define SPIFLASH_ENWRITESTATUSREG	0x50 // Enable write status register
#define SPIFLASH_WRITESTATUSREG	0x01 // write status register
#define SPIFLASH_ENWRITE	0x06 // write enable
#define SPIFLASH_DISWRITE	0x04 // write disable
#define SPIFLASH_READID	0x90 // read chip ID

#define	spi_CS_en()	PORTB &= ~(1<<SPI_SS_PIN); _delay_us(200);
#define	spi_CS_dis()	_delay_us(200); PORTB |= (1<<SPI_SS_PIN);

#define spiflash_BP(x) (spiflash_writestatusreg(((x) & 0x3)<<2))
#define spiflash_isbusy() (spiflash_readstatusreg() & 0x1)
#define spiflash_iswriteen() ((spiflash_readstatusreg() & 0x2)>>1)


// setup spi
void setup_spi(uint8_t mode,   // timing mode SPI_MODE[0-4]
	       int dord,             // data direction SPI_LSB|SPI_MSB
	       int interrupt,        // whether to raise interrupt on receive
	       uint8_t clock); // clock divisor

// disable spi
void disable_spi(void);

// send and receive a byte of data (master mode)
uint8_t trans_spi(uint8_t out);
uint8_t spiflash_readstatusreg(void);
void spiflash_enwrite(void);
void spiflash_diswrite(void);
void spiflash_writestatusreg(uint8_t data);
void spiflash_readblock(uint8_t *data, uint32_t address, uint8_t count);
void spiflash_writeblock(uint8_t *data, uint32_t address, uint8_t count);
void spiflash_writebyte(uint32_t address, uint8_t data);
uint16_t spiflash_read_id(void);
void spiflash_sector_erase(uint32_t address);
void spiflash_block_erase(uint32_t address);
void spiflash_chip_erase(void);

#ifdef __cplusplus
} // extern "C"
#endif

#endif

